const Session = require('./Session');
const KeyspaceService = require('./service/Keyspace/KeyspaceService');
const messages = require("./service/Session/autogenerated/Session_pb");

/**
 * Entry-point for Grakn client, it communicates with a running Grakn server using gRPC.
 * It allows to:
 * - execute operations on keyspaces
 * - obtain a new Session bound to a specific keyspace
 * 
 * @param {String} uri String containing host address and gRPC port of a running Grakn instance, e.g. "localhost:48555"
 * @param {Object} credentials Optional object containing user credentials - only used when connecting to a KGMS instance
 */
function Grakn(uri, credentials) {
    const keyspaceService = new KeyspaceService(uri, credentials);

    this.session = (keyspace) => new Session(uri, keyspace, credentials);

    this.keyspace = {
        delete: (keyspace) => keyspaceService.delete(keyspace),
        retrieve: () => keyspaceService.retrieve()
    };
}

module.exports = Grakn

/**
 * List of available dataTypes for Grakn Attributes
 */
module.exports.dataType = {
    STRING: messages.AttributeType.DATA_TYPE.STRING,
    BOOLEAN: messages.AttributeType.DATA_TYPE.BOOLEAN,
    INTEGER: messages.AttributeType.DATA_TYPE.INTEGER,
    LONG: messages.AttributeType.DATA_TYPE.LONG,
    FLOAT: messages.AttributeType.DATA_TYPE.FLOAT,
    DOUBLE: messages.AttributeType.DATA_TYPE.DOUBLE,
    DATE: messages.AttributeType.DATA_TYPE.DATE
};

/**
 * List of available transaction types supported by Grakn
 */
module.exports.txType = {
    READ: messages.Transaction.Type.READ,
    WRITE: messages.Transaction.Type.WRITE,
    BATCH: messages.Transaction.Type.BATCH
};